/*-
 * Copyright (C) @BABOLO  2002 Dec 04
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE AUTHOR AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE AUTHOR OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 * $Id: pgobin.c,v 1.1 2002/12/10 02:28:52 babolo Exp $
 */

#ifndef lint
char copyright[] = "\
@(#)Copyright (C) @BABOLO  2002 Dec 04\n\
@(#)All rights reserved.\n";
static char rcsid[] = "$Id: pgobin.c,v 1.1 2002/12/10 02:28:52 babolo Exp $";
#endif /* not lint */
#include <sys/types.h>
#include <sysexits.h>
#include <unistd.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>
#include <stdio.h>
#include <err.h>
#include <libpq-fe.h>
#define BUFSIZE 8192
#define PQEXEC(RESULT, QUERY, CONDITION) {                                                           \
    if  (v_flag > 2) fprintf(stderr, "PQexec: %s\n", QUERY);                                         \
    RESULT = PQexec(options->conn, QUERY);                                                           \
    if  (!RESULT) {                                                                                  \
        fprintf(stderr, "result is absent - fatal error.\n");                                        \
        exit(EX_DATAERR);                                                                            \
    }                                                                                                \
    if  (v_flag > 2) fprintf(stderr, "PQresultStatus %d\n", PQresultStatus(res));                    \
    if  (PQresultStatus(res) != CONDITION) {                                                         \
        fprintf(stderr, "PQexec failed: %s\n", PQresultErrorMessage(RESULT));                        \
        exit(EX_DATAERR);                                                                            \
}   }

int v_flag;

struct connopt{
    char   *dbname;
    char   *host;
    char   *port;
    char   *username;
    PGconn *conn;
    char   *query;
};

void
usage(int ex) {
    fprintf(stderr, "pgobin " VERS "  " DATE "
Usage: pgobin -[dhpqUv] [query file]
    -d DataBase name
    -h Hostname of backend server
    -p Port
    -U Username
    -v Debug print
");
    exit(ex);
}

int
main(int argc, char *argv[]) {
    struct connopt *options = malloc(sizeof(struct connopt));
    FILE *in = stdin;
    FILE *out = stdout;
    int c;
    size_t i, j;
    PGresult *res;

    if  (options == NULL)                                   errx(EX_UNAVAILABLE, "Malloc failed #1\n");
    options->dbname = NULL;
    options->host = NULL;
    options->port = NULL;
    options->username = NULL;
    options->query = NULL;
    v_flag = 0;  /*   */
    while ((c = getopt(argc, argv, "d:h:p:qU:v?")) != -1) {
        switch (c) {
        case 'd': options->dbname = optarg;
                  break;
        case 'h': options->host = optarg;
                  break;
        case 'p': options->port = optarg;
                  break;
        case 'q': v_flag--;
                  break;
        case 'U': options->username = optarg;
                  break;
        case 'v': v_flag++;
                  break;
        case '?': usage(EX_OK);
        default:  usage(EX_USAGE);
    }   }
    argc -= optind;
    argv += optind;
    if  (argc > 0 && argv && *argv) {
        in = fopen(*argv, "r");
        if  (in == NULL)                                            err(EX_NOINPUT, "file %s:", *argv);
    }

    /* Prepare the query */
    if  ((options->query = malloc(j = BUFSIZE)) == NULL)    errx(EX_UNAVAILABLE, "Malloc failed #1\n");
    strncpy(options->query, "DECLARE pgobin BINARY CURSOR FOR ", BUFSIZE - 1);
    for (i = strlen(options->query); !feof(in) && !ferror(in);) {
        if  (v_flag > 2) fprintf(stderr, "i=%d j=%d eof=%d err=%d\n", i, j, feof(in), ferror(in));
        if  (i + 1 >= j && (options->query = realloc(options->query, j += BUFSIZE)) == NULL)
                                                            errx(EX_UNAVAILABLE, "Malloc failed #2\n");
        if  (v_flag > 2) fprintf(stderr, "i=%d, j=%d\n", i, j);
        i += fread(&(options->query[i]), 1, j - i - 1, in);
        options->query[i] = '\0';
        if  (v_flag > 2) fprintf(stderr, "%s~\n", options->query);
    }
    if  (ferror(in)) {
        fprintf(stderr, "Input error\n");
        exit(EX_IOERR);
    }

    /* Connect to DBMS */
    options->conn = PQsetdbLogin( options->host
                                , options->port
                                , NULL, NULL
                                , options->dbname
                                , options->username
                                , NULL
                                );
    if  (v_flag > 3) fprintf(stderr, "PQstatus %d\n", PQstatus(options->conn));
    if  (PQstatus(options->conn) == CONNECTION_BAD) {
        if  (v_flag >= 0)
               fprintf(stderr, "Connection to PostgreSQL failed: %s\n", PQerrorMessage(options->conn));
        exit(EX_DATAERR);
    }

    PQEXEC(res, "BEGIN", PGRES_COMMAND_OK);
    PQclear(res);
    PQEXEC(res, options->query, PGRES_COMMAND_OK);
    PQclear(res);
    PQEXEC(res, "FETCH FROM pgobin", PGRES_TUPLES_OK);
    for (; PQntuples(res) > 0 && !feof(out) && !ferror(out); ) {
        for (c = 0; c < PQnfields(res) && !feof(out) && !ferror(out); c++) {
            if  (!PQgetisnull(res, 0, c)) {
                fwrite(PQgetvalue(res, 0, c), PQgetlength(res, 0, c), 1, out);
        }   }
        PQclear(res);
        PQEXEC(res, "FETCH FROM pgobin", PGRES_TUPLES_OK)
    }
    PQEXEC(res, "END", PGRES_COMMAND_OK);
    PQclear(res);
    PQfinish(options->conn);
    exit(EX_OK);
}
